;; Generated package description from org.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "org" "9.8.5" "Outline-based notes management and organizer" '((emacs "28.2")) :commit "0d02d72627428dddf21ec66b1eb87e46c2849ded" :authors '(("Carsten Dominik" . "carsten.dominik@gmail.com")) :maintainer '("Ihor Radchenko" . "yantar92@posteo.net") :keywords '("outlines" "hypermedia" "calendar" "text") :url "https://orgmode.org")
